from django.http import Http404, HttpResponseForbidden
from django.shortcuts import get_object_or_404, redirect, render

from flags.models import FlagState
from flags.sources import get_flags
from flags.templatetags.flags_debug import bool_enabled

from wagtailflags.forms import FlagStateForm, NewFlagForm
from wagtailflags.signals import flag_disabled, flag_enabled
from wagtailflags.templatetags.wagtailflags_admin import deletable


def index(request):
    context = {
        "flags": sorted(get_flags().values(), key=lambda x: x.name),
    }
    return render(request, "wagtailflags/index.html", context)


def create_flag(request):
    """Create a new flag.
    This will add a FlagState object with a custom name and a boolean: False
    condition"""
    if request.method == "POST":
        form = NewFlagForm(
            request.POST,
            initial={
                "condition": "path matches",
                "value": "/foo",
                "required": False,
            },
        )
        if form.is_valid():
            form.save()
            return redirect("wagtailflags:flag_index", name=form.instance.name)
    else:
        form = NewFlagForm()

    context = dict(form=form)
    return render(request, "wagtailflags/flags/create_flag.html", context)


def delete_flag(request, name):
    """Delete a database flag."""
    flag = get_flags().get(name)

    if not flag:
        raise Http404

    if not deletable(flag):
        return HttpResponseForbidden(name)

    if request.method == "POST":
        FlagState.objects.filter(name=name).delete()
        return redirect("wagtailflags:list")

    context = {
        "flag": flag,
    }
    return render(request, "wagtailflags/flags/delete_flag.html", context)


def flag_index(request, name):
    flag = get_flags().get(name)

    if not flag:
        raise Http404

    # If there's a database boolean condition, fetch it and treat it as a
    # on/off switch
    if "enable" in request.GET or "disable" in request.GET:
        db_boolean_condition = next(
            (
                c
                for c in flag.conditions
                if c.condition == "boolean"
                and getattr(c, "obj", None) is not None
            ),
            None,
        )

        if db_boolean_condition is None:
            boolean_condition_obj = FlagState.objects.create(
                name=name, condition="boolean", value="True"
            )
        else:
            boolean_condition_obj = db_boolean_condition.obj

        if "enable" in request.GET and not bool_enabled(flag):
            boolean_condition_obj.value = True
            flag_enabled.send(sender=flag_index, flag_name=flag.name)
        elif "disable" in request.GET and bool_enabled(flag):
            boolean_condition_obj.value = False
            flag_disabled.send(sender=flag_index, flag_name=flag.name)

        boolean_condition_obj.save()
        return redirect("wagtailflags:flag_index", name=name)

    context = {
        "flag": flag,
    }
    return render(request, "wagtailflags/flags/flag_index.html", context)


def edit_condition(request, name, condition_pk=None):
    flag = get_flags().get(name)

    if not flag:
        raise Http404

    try:
        condition = FlagState.objects.get(pk=condition_pk)
        title_str = f"Edit {condition} condition"
    except FlagState.DoesNotExist:
        condition = None
        title_str = f"Create a condition on {flag.name}"

    if request.method == "POST":
        form = FlagStateForm(
            request.POST, initial={"name": name}, instance=condition
        )
        if form.is_valid():
            form.save()
            return redirect("wagtailflags:flag_index", name=name)

    else:
        form = FlagStateForm(initial={"name": name}, instance=condition)

    context = {
        "flag": flag,
        "form": form,
        "condition_str": str(condition),
        "condition_pk": condition_pk,
        "title": title_str,
    }
    return render(request, "wagtailflags/flags/edit_condition.html", context)


def delete_condition(request, name, condition_pk):
    flag = get_flags().get(name)

    if not flag:
        raise Http404

    condition = get_object_or_404(FlagState, pk=condition_pk)

    if request.method == "POST":
        condition.delete()
        return redirect("wagtailflags:flag_index", name=name)

    context = {
        "flag": flag,
        "condition_str": str(condition),
        "condition_pk": condition.pk,
    }
    return render(request, "wagtailflags/flags/delete_condition.html", context)



from django.test import TestCase

from wagtail.test.utils import WagtailTestUtils

from flags.models import FlagState


class TestWagtailFlagsViews(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.dbonly_flag = FlagState.objects.create(
            name="DBONLY_FLAG",
            condition="boolean",
            value="True",
            required=False,
        )

    def test_flags_index(self):
        response = self.client.get("/admin/flags/")
        self.assertEqual(response.status_code, 200)

        self.assertContains(response, "FLAG_ENABLED")
        self.assertContains(response, "FLAG_DISABLED")
        self.assertContains(response, "DBONLY_FLAG")
        self.assertContains(response, "<b>enabled</b> when")
        self.assertContains(response, "<b>enabled</b> for")

    def test_flag_create(self):
        response = self.client.get("/admin/flags/create/")
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            "/admin/flags/create/", {"name": "NEW_FLAG"}
        )
        self.assertRedirects(response, "/admin/flags/NEW_FLAG/")
        self.assertEqual(FlagState.objects.count(), 2)
        new_flag_condition = FlagState.objects.get(name="NEW_FLAG")
        self.assertEqual(new_flag_condition.condition, "boolean")
        self.assertEqual(new_flag_condition.value, "False")
        self.assertEqual(new_flag_condition.required, False)

    def test_flag_create_existing(self):
        response = self.client.get("/admin/flags/create/")
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            "/admin/flags/create/", {"name": "DBONLY_FLAG"}
        )
        self.assertContains(response, "Flag named DBONLY_FLAG already exists")

    def test_flag_delete(self):
        response = self.client.get("/admin/flags/DBONLY_FLAG/delete/")
        self.assertEqual(response.status_code, 200)

        response = self.client.post("/admin/flags/DBONLY_FLAG/delete/")

        self.assertRedirects(response, "/admin/flags/")
        self.assertEqual(FlagState.objects.count(), 0)

    def test_flag_delete_nonexistent(self):
        response = self.client.get(
            "/admin/flags/THIS_FLAG_DOES_NOT_EXIST/delete/"
        )
        self.assertEqual(response.status_code, 404)

    def test_flag_delete_not_deletable(self):
        response = self.client.get("/admin/flags/FLAG_ENABLED/delete/")
        self.assertEqual(response.status_code, 403)

    def test_flag_index_nonexistent_flag_raises_404(self):
        response = self.client.get("/admin/flags/THIS_FLAG_DOES_NOT_EXIST/")
        self.assertEqual(response.status_code, 404)

    def test_flag_index(self):
        response = self.client.get("/admin/flags/FLAG_DISABLED/")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "FLAG_DISABLED")
        self.assertContains(response, "path matches")
        self.assertContains(response, "Enable FLAG_DISABLED")

    def test_flag_index_enabled(self):
        response = self.client.get("/admin/flags/FLAG_ENABLED/")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Disable FLAG_ENABLED")

    def test_flag_index_disabled(self):
        response = self.client.get("/admin/flags/FLAG_DISABLED/")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Enable FLAG_DISABLED")

    def test_flag_index_enabled_required_true(self):
        response = self.client.get("/admin/flags/FLAG_ENABLED/")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Disable FLAG_ENABLED")

    def test_flag_index_disabled_required_true(self):
        response = self.client.get("/admin/flags/FLAG_ENABLED/")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Disable FLAG_ENABLED")

    def test_enable_flag(self):
        condition_query = FlagState.objects.filter(name="FLAG_DISABLED")
        self.assertEqual(len(condition_query.all()), 0)

        response = self.client.get(
            "/admin/flags/FLAG_DISABLED/", {"enable": ""}
        )
        self.assertRedirects(response, "/admin/flags/FLAG_DISABLED/")
        self.assertEqual(len(condition_query.all()), 1)
        self.assertEqual(condition_query.first().condition, "boolean")
        self.assertEqual(condition_query.first().value, "True")

    def test_enable_flag_with_required_true(self):
        condition_query = FlagState.objects.filter(name="FLAG_DISABLED")
        self.assertEqual(len(condition_query.all()), 0)

        self.client.get("/admin/flags/FLAG_DISABLED/", {"enable": ""})
        self.assertEqual(len(condition_query.all()), 1)
        self.assertEqual(condition_query.first().condition, "boolean")
        self.assertEqual(condition_query.first().value, "True")

    def test_disable_flag(self):
        condition_query = FlagState.objects.filter(name="DBONLY_FLAG")
        self.client.get("/admin/flags/DBONLY_FLAG/", {"disable": ""})
        self.assertEqual(len(condition_query.all()), 1)
        self.assertEqual(condition_query.first().condition, "boolean")
        self.assertEqual(condition_query.first().value, "False")

    def test_disable_flag_with_required_true(self):
        condition_query = FlagState.objects.filter(name="DBONLY_FLAG")
        self.client.get("/admin/flags/DBONLY_FLAG/", {"disable": ""})
        self.assertEqual(len(condition_query.all()), 1)
        self.assertEqual(condition_query.first().condition, "boolean")
        self.assertEqual(condition_query.first().value, "False")

    def test_create_flag_condition_nonexistent_flag_raises_404(self):
        response = self.client.get(
            "/admin/flags/THIS_FLAG_DOES_NOT_EXIST/create/"
        )
        self.assertEqual(response.status_code, 404)

    def test_create_flag_condition(self):
        response = self.client.get("/admin/flags/DBONLY_FLAG/create/")
        self.assertEqual(response.status_code, 200)

        params = {
            "condition": "path matches",
            "value": "/db_path",
        }
        response = self.client.post("/admin/flags/DBONLY_FLAG/create/", params)
        self.assertRedirects(response, "/admin/flags/DBONLY_FLAG/")
        self.assertEqual(len(FlagState.objects.all()), 2)

    def test_edit_flag_condition_nonexistent_flag_raises_404(self):
        response = self.client.get("/admin/flags/THIS_FLAG_DOES_NOT_EXIST/99/")
        self.assertEqual(response.status_code, 404)

    def test_edit_flag_condition(self):
        condition_obj = FlagState.objects.create(
            name="DBONLY_FLAG", condition="boolean", value="true"
        )
        response = self.client.get(
            "/admin/flags/DBONLY_FLAG/{}/".format(condition_obj.pk)
        )
        self.assertEqual(response.status_code, 200)

        params = {
            "condition": "boolean",
            "value": "true",
        }

        response = self.client.post(
            "/admin/flags/DBONLY_FLAG/{}/".format(condition_obj.pk), params
        )
        self.assertRedirects(response, "/admin/flags/DBONLY_FLAG/")
        self.assertEqual(
            FlagState.objects.get(pk=condition_obj.pk).value, "true"
        )

    def test_delete_flag_condition_nonexistent_flag_raises_404(self):
        response = self.client.get(
            "/admin/flags/THIS_FLAG_DOES_NOT_EXIST/99/delete/"
        )
        self.assertEqual(response.status_code, 404)

    def test_delete_flag_condition(self):
        condition_obj = FlagState.objects.create(
            name="DBONLY_FLAG", condition="boolean", value="true"
        )
        self.assertEqual(len(FlagState.objects.all()), 2)
        response = self.client.get(
            "/admin/flags/DBONLY_FLAG/{}/delete/".format(condition_obj.pk)
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            "/admin/flags/DBONLY_FLAG/{}/delete/".format(condition_obj.pk)
        )
        self.assertRedirects(response, "/admin/flags/DBONLY_FLAG/")
        self.assertEqual(len(FlagState.objects.all()), 1)



from django import forms

from flags.forms import FlagStateForm as DjangoFlagsFlagStateForm
from flags.models import FlagState
from flags.sources import get_flags


class NewFlagForm(forms.ModelForm):
    name = forms.CharField(label="Name", required=True)

    def clean_name(self):
        name = self.cleaned_data["name"]
        if name in get_flags():
            raise forms.ValidationError(
                "Flag named {} already exists".format(name)
            )
        return name

    def save(self, commit=True):
        obj = super(NewFlagForm, self).save(commit=False)
        obj.condition = "boolean"
        obj.value = "False"
        obj.required = False
        obj.save()
        return obj

    class Meta:
        model = FlagState
        fields = ("name",)


class FlagStateForm(DjangoFlagsFlagStateForm):
    name = forms.CharField(
        label="Flag",
        required=True,
        disabled=True,
        widget=forms.HiddenInput(),
    )

    class Meta:
        model = FlagState
        fields = ("name", "condition", "value", "required")



from django import template

from flags.sources import DatabaseCondition
from flags.templatetags.flags_debug import bool_enabled


register = template.Library()


@register.filter
def enablable(flag):
    """Return true if a flag is enablable by Wagtail-Flags.

    A flag is enablable by Wagtail-Flags if it has a required boolean
    condition and that boolean condition is False.
    """
    return not any(
        c.required for c in flag.conditions if c.condition == "boolean"
    ) and not bool_enabled(flag)


@register.filter
def disablable(flag):
    """Return true if a flag is disablable by Wagtail-Flags.

    A flag is disablable by Wagtail-Flags if it has a required boolean
    condition and that boolean condition is True.
    """
    return not any(
        c.required for c in flag.conditions if c.condition == "boolean"
    ) and bool_enabled(flag)


@register.filter
def deletable(flag):
    """Return true if a flag is deletable by Wagtail-Flags.

    A flag is deletable by Wagtail-Flags if it is database-only, which means
    all of its conditions are type DatabaseCondition.
    """
    return flag.conditions and all(
        isinstance(c, DatabaseCondition) for c in flag.conditions
    )



from unittest import mock

from django.test import TestCase

from wagtail.test.utils import WagtailTestUtils

from flags.models import FlagState

from wagtailflags.signals import flag_disabled, flag_enabled


class SignalsTestCase(TestCase, WagtailTestUtils):
    def setUp(self):
        self.login()

        self.dbonly_flag = FlagState.objects.create(
            name="DBONLY_FLAG",
            condition="boolean",
            value="True",
            required=False,
        )

    def test_flag_enabled_was_sent(self):
        mock_handler = mock.Mock()
        flag_enabled.connect(mock_handler)
        self.client.get("/admin/flags/FLAG_DISABLED/", {"enable": ""})
        mock_handler.assert_called_with(
            sender=mock.ANY, flag_name="FLAG_DISABLED", signal=mock.ANY
        )

    def test_flag_disabled_was_sent(self):
        mock_handler = mock.Mock()
        flag_disabled.connect(mock_handler)
        self.client.get("/admin/flags/DBONLY_FLAG/", {"disable": ""})
        mock_handler.assert_called_with(
            sender=mock.ANY, flag_name="DBONLY_FLAG", signal=mock.ANY
        )



from django.test import TestCase, override_settings

from flags.models import FlagState
from flags.sources import get_flags

from wagtailflags.templatetags.wagtailflags_admin import (
    deletable,
    disablable,
    enablable,
)


class TestWagtailFlagsAdminTemplateTags(TestCase):
    @override_settings(FLAGS={"MYFLAG": [("boolean", False)]})
    def test_enablable_disabled_not_required(self):
        flag = get_flags().get("MYFLAG")
        self.assertTrue(enablable(flag))

    @override_settings(FLAGS={"MYFLAG": [("boolean", True)]})
    def test_enablable_enabled_not_required(self):
        flag = get_flags().get("MYFLAG")
        self.assertFalse(enablable(flag))

    @override_settings(FLAGS={"MYFLAG": [("boolean", True, True)]})
    def test_enablable_disabled_required(self):
        flag = get_flags().get("MYFLAG")
        self.assertFalse(enablable(flag))

    @override_settings(FLAGS={"MYFLAG": [("boolean", True)]})
    def test_disablable_disabled_not_required(self):
        flag = get_flags().get("MYFLAG")
        self.assertTrue(disablable(flag))

    @override_settings(FLAGS={"MYFLAG": [("boolean", False)]})
    def test_disablable_enabled_not_required(self):
        flag = get_flags().get("MYFLAG")
        self.assertFalse(disablable(flag))

    @override_settings(FLAGS={"MYFLAG": [("boolean", False, True)]})
    def test_disablable_enabled_required(self):
        flag = get_flags().get("MYFLAG")
        self.assertFalse(disablable(flag))

    @override_settings(FLAGS={"MYFLAG": [("boolean", False)], "EMPTYFLAG": []})
    def test_deletable(self):
        FlagState.objects.create(
            name="DBFLAG",
            condition="boolean",
            value="True",
        )
        self.assertFalse(deletable(get_flags().get("MYFLAG")))
        self.assertFalse(deletable(get_flags().get("EMPTYFLAG")))
        self.assertTrue(deletable(get_flags().get("DBFLAG")))



from django.test import RequestFactory, TestCase

from wagtail.models import Site

from flags.conditions import RequiredForCondition

from wagtailflags.conditions import site_condition


class SiteConditionTestCase(TestCase):
    def setUp(self):
        self.site = Site.objects.get(is_default_site=True)
        self.factory = RequestFactory()
        self.request = self.factory.get("/")
        Site.find_for_request(self.request)

    def test_site_valid_string(self):
        self.assertTrue(site_condition("localhost:80", request=self.request))

    def test_site_valid_string_no_port(self):
        self.assertTrue(site_condition("localhost", request=self.request))

    def test_site_valid_string_default_port(self):
        self.assertTrue(
            site_condition("localhost [default]", request=self.request)
        )

    def test_site_valid_site(self):
        self.assertTrue(site_condition(str(self.site), request=self.request))

    def test_site_invalid_site(self):
        self.assertFalse(
            site_condition("non.existent.site", request=self.request)
        )

    def test_request_required(self):
        with self.assertRaises(RequiredForCondition):
            site_condition("localhost:80")



import django
from django.templatetags.static import static
from django.urls import include, re_path, reverse
from django.utils.html import format_html

from wagtail import hooks
from wagtail.admin.menu import MenuItem

from wagtailflags import views


@hooks.register("register_settings_menu_item")
def register_flags_menu():
    return MenuItem(
        "Flags",
        reverse("wagtailflags:list"),
        icon_name="flag",
        order=10000,
    )


@hooks.register("register_admin_urls")
def register_flag_admin_urls():
    flagpatterns = [
        re_path(r"^$", views.index, name="list"),
        re_path(r"^create/$", views.create_flag, name="create_flag"),
        re_path(r"^(?P<name>[\w\-]+)/$", views.flag_index, name="flag_index"),
        re_path(
            r"^(?P<name>[\w\-]+)/delete/$",
            views.delete_flag,
            name="delete_flag",
        ),
        re_path(
            r"^(?P<name>[\w\-]+)/create/$",
            views.edit_condition,
            name="create_condition",
        ),
        re_path(
            r"^(?P<name>[\w\-]+)/(?P<condition_pk>\d+)/$",
            views.edit_condition,
            name="edit_condition",
        ),
        re_path(
            r"^(?P<name>[\w\-]+)/(?P<condition_pk>\d+)/delete/$",
            views.delete_condition,
            name="delete_condition",
        ),
    ]

    if django.VERSION >= (1, 10):  # pragma: no cover
        urlpatterns = [
            re_path(
                r"^flags/",
                include(
                    (flagpatterns, "wagtailflags"), namespace="wagtailflags"
                ),
            )
        ]
    else:  # pragma: no cover; fallback for Django < 1.10
        urlpatterns = [
            re_path(
                r"^flags/",
                include((flagpatterns, "wagtailflags", "wagtailflags")),
            )
        ]

    return urlpatterns


@hooks.register("insert_global_admin_css", order=100)
def global_admin_css():
    return format_html(
        '<link rel="stylesheet" href="{}">',
        static("wagtailflags/css/wagtailflags.css"),
    )


@hooks.register("register_icons")
def register_icons(icons):
    return icons + ["wagtailflags/flag.svg"]
